import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import pickle
import math

from dataclasses import dataclass, field

from scipy.constants import elementary_charge, hbar, electron_mass, physical_constants, speed_of_light

from scipy.sparse.linalg import expm, expm_multiply
from scipy.sparse import save_npz, load_npz, csr_array, csc_array, diags, kron, kronsum, identity
from scipy.sparse.linalg import lobpcg, LinearOperator, inv, norm

iterator = 0

@dataclass(frozen=True)
class ph_con():
    mu_b : float = physical_constants['Bohr magneton in eV/T'][0]
    hb : float = hbar
    m0 : float = electron_mass
    ec : float = elementary_charge
    c : float = speed_of_light
    mhz2ev : float = 4.13566553853598*1e-9
@dataclass(frozen=True)
class res():
    n_photon : int = 3
    k_tot : float = 2.5 # MHz
    k_int : float = 1.5 # Mhz
    f_res : float = 6910 # MHz
    f_read : float = 6907.6 # MHz

@dataclass(frozen=True)
class dot_1():
    
    gamma_rel_s : float = 1/0.178 # MHz
    gamma_dep_s : float = 1/0.07 # MHz
    
    
    leverarm : float = 30*1e-3 # eV/V
@dataclass(frozen=True)
class dot_2():

    
    gamma_rel_s : float = 1/0.26 # MHz
    gamma_dep_s : float = 1/0.05 # MHz
    
    leverarm : float = 30*1e-3 # eV/V


entries : np.ndarray = np.array( [np.sqrt(it+1) for it in range(0,res.n_photon-1)] , dtype = float)

@dataclass(frozen=True)
class operators():
    i : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,1]],dtype = complex))
    x : np.ndarray = field(default_factory = lambda :np.array([[0,1],[1,0]],dtype = complex))
    y : np.ndarray = field(default_factory = lambda :np.array([[0,-1j],[1j,0]],dtype = complex))
    z : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,-1]],dtype = complex))
    pl : np.ndarray = field(default_factory = lambda :np.array([[0,1],[0,0]],dtype = complex))
    mi : np.ndarray = field(default_factory = lambda :np.array([[0,0],[1,0]],dtype = complex))
    pl_star : np.ndarray = field(default_factory = lambda :np.array([[1,1],[-1,-1]],dtype = complex)/2)
    mi_star : np.ndarray = field(default_factory = lambda :np.array([[1,-1],[1,-1]],dtype = complex)/2)
    p0 : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,0]],dtype = complex))
    p1 : np.ndarray = field(default_factory = lambda :np.array([[0,0],[0,1]],dtype = complex))
    
    ii : np.ndarray = field(default_factory = lambda :np.identity(4,dtype = complex))
    i4 : np.ndarray = field(default_factory = lambda :np.identity(4**2,dtype = complex))
    
    
    a : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1))
    ad : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1).T)
    ada : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1).T @ np.diag(entries,k=1))
    ire : np.ndarray = field(default_factory = lambda :np.identity(res.n_photon,dtype = complex))
    
    res0 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
    # res0[0,0] =1
    if res.n_photon > 1:
        res1 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res1[1,1] =1
    if res.n_photon > 2:
        res2 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res2[2,2] =1
    if res.n_photon > 3:
        res3 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res3[3,3] =1
    if res.n_photon > 4:
        res4 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res4[4,4] =1
    if res.n_photon > 5:
        res5 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res5[5,5] =1
    if res.n_photon > 6:
        res6 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res6[6,6] =1

pauli = dict({'i':np.array([[1,0],[0,1]],dtype = complex),
              'x':np.array([[0,1],[1,0]],dtype = complex),
              'y':np.array([[0,-1j],[1j,0]],dtype = complex),
              'z':np.array([[1,0],[0,-1]],dtype = complex),
              '+':np.array([[0,1],[0,0]],dtype = complex),
              '-':np.array([[0,0],[1,0]],dtype = complex),
              '+*':np.array([[1,1],[-1,-1]],dtype = complex)/2,
              '-*':np.array([[1,-1],[1,-1]],dtype = complex)/2,
              '0':np.array([[1,0],[0,0]],dtype = complex),
              '1':np.array([[0,0],[0,1]],dtype = complex),
              })

pauli['x+'] = (pauli['i']+pauli['x'])/2
pauli['x-'] = (pauli['i']-pauli['x'])/2
pauli['y+'] = (pauli['i']+pauli['y'])/2
pauli['y-'] = (pauli['i']-pauli['y'])/2
pauli['z+'] = (pauli['i']+pauli['z'])/2
pauli['z-'] = (pauli['i']-pauli['z'])/2


def compute_eigenvalues(ham, return_vecs : bool = False):
    # returns eigenvalues and optionally eigenvectors
    eig_val_s, eig_vec = np.linalg.eig(ham)
    idx = eig_val_s.argsort()
    eig_val = eig_val_s[idx]
    if return_vecs is False:
        return eig_val/(2*np.pi)
    else:
        eig_vec = eig_vec[:, idx].T
        return eig_val/(2*np.pi) , eig_vec
    

def operator_disp(spin_1 : str = 'i', spin_2 : str = 'i') -> np.ndarray:
    op_s1 = pauli[spin_1]
    op_s2 = pauli[spin_2]
    
    matrix = np.kron(op_s1,op_s2)
    
    return matrix

def fitting_iswap_dispersive(t_array : np.ndarray, 
                  f_qubit_dot_1 : float = 6779.2, # total resonance frequency minus dispersive shift
                  f_qubit_dot_2 : float = 6779.2, # total resonance frequency minus dispersive shift
                  exchange : float = 19., # total swap rate
                  disp_shift_dot_1 : float = 20. , # dispersive shift dot 1
                  disp_shift_dot_2 : float = 20. , # dispersive shift dot 2
                  g1_dot_1_spin : float = dot_1.gamma_rel_s, # spin-relaxation dot 1
                  g2_dot_1_spin : float = dot_1.gamma_dep_s, # spin-dephasing dot 1
                  g1_dot_2_spin : float = dot_2.gamma_rel_s, # spin-relaxation dot 2
                  g2_dot_2_spin : float = dot_2.gamma_dep_s, # spin-dephasing dot 2
                  kappa_res : float = res.k_tot,
                  result : str = "expectation", # type of result: 
                                               #'expectation' normalized expectation values, 
                                               #'resonator' normalized readout values
                  q_excite: int = 1
                  ):
    # all energies in MHz and set hbar = 1
    
    detuning = 110 # spin-resonator detuning during the readout phase
    
    # Set up the Hamiltonian without coupling to get initial states
    hamiltonian_static = 2*np.pi*(
        # zero order terms
        - operator_disp(spin_1 = 'z') * f_qubit_dot_1/2
        - operator_disp(spin_2 = 'z') * f_qubit_dot_2/2
        )
        # swapping terms
    hamiltonian_swap = 2*np.pi*(
        operator_disp(spin_1 = 'x',spin_2 = 'x') 
        + operator_disp(spin_1 = 'y',spin_2 = 'y')
        )*exchange/4
        
    # Set up time-steps. Currently only uniform time traces are supported
    t_array = t_array
    t_step = t_array[1]-t_array[0]
    # Set up the initial states from diagonalizing the non-interacting Hamiltonian
    t_eigs, t_vecs = compute_eigenvalues(hamiltonian_static, return_vecs = True)
    init_DQD1_ex = t_vecs[2]
    init_DQD2_ex = t_vecs[1]
    # Set up the interacting Hamiltonian for both dots
    hamiltonian = (hamiltonian_swap + hamiltonian_static)
    
    idmatrix = np.identity(len(hamiltonian), dtype = complex)
    
    # Set up Liuovillian for the system
    liuoville = -1j*(np.kron(hamiltonian,idmatrix)
                            -np.kron(idmatrix,hamiltonian.T))
    
    
    # set up all Lindblad terms
    
    lindblad_1 = np.sqrt(g2_dot_1_spin/2) * operator_disp(spin_1 = 'z')
    lindblad_2 = np.sqrt(g2_dot_2_spin/2) * operator_disp(spin_2 = 'z')
    
    lindblad_3 = np.sqrt(g1_dot_1_spin) * operator_disp(spin_1 = '+')
    lindblad_4 = np.sqrt(g1_dot_2_spin) * operator_disp(spin_2 = '+')
    
    
    # create lindbladadian
    lindblad = np.zeros(liuoville.shape,dtype=complex)
    
    list_lindblads = [lindblad_1,lindblad_2,lindblad_3,
                      lindblad_4]
    
    # add Lindblad dissipators
    
    for dissipator in list_lindblads:
        lindblad += 1/2*(2*np.kron(dissipator,dissipator)
                          -np.kron(idmatrix,(dissipator.T @ dissipator).T)
                          -np.kron((dissipator.T @ dissipator),idmatrix))
    
    # create initial dm states
    # init_dm = [np.outer(dm_a, dm_a.conj()) 
    #            for dm_a in [init_DQD1,init_DQD2]]
    
    init_dm = [np.outer(dm_a, dm_a.conj())
               for dm_a in [init_DQD1_ex,init_DQD2_ex]]
    
    
    # vectorize initial dm states
    init_vec = [dm.flatten() for dm in init_dm]
    # init_vec = [init_vec_raw[it]*1.+0.*(dm).flatten() for it,dm in enumerate(init_dm)]
    
    # Set up full system 
    lindbladian = lindblad + liuoville
    # Set up full system for a time step
    lindbladian_t = lindbladian*t_step
    
    # Set up state for each time step
    final_vec_01 = [init_vec[0]]
    final_vec_10 = [init_vec[1]]
    
    # Compute state for each time step by serial propagating it with lindbladian_t
    for iterator in range(len(t_array)-1):
        final_vec_01.append(sp.linalg.expm(lindbladian_t) @ final_vec_01[-1])
        final_vec_10.append(sp.linalg.expm(lindbladian_t) @final_vec_10[-1])
    
    
    # Set up readout. Theory can be found in 
    # Kohler et al., 10.1103/PhysRevA.98.023849
    
    # Compute dispersive shift of final state
    
    
    # Simulate measurement using general input-output formalism
    measurement_DQD_1_ex_1_read = []
    measurement_DQD_1_ex_2_read  = []
    measurement_DQD_2_ex_1_read  = []
    measurement_DQD_2_ex_2_read = []
    for iterator in range(len(t_array)):
        
        output_density_01 = np.reshape(final_vec_01[iterator],[4,4])
        output_density_10 = np.reshape(final_vec_10[iterator],[4,4])
        
        expect_DQD_1_ex_1_read = np.trace(-operator_disp(spin_1 = 'z') @ output_density_01)
        expect_DQD_1_ex_2_read = np.trace(-operator_disp(spin_2 = 'z') @ output_density_01)
        
        expect_DQD_2_ex_1_read = np.trace(-operator_disp(spin_1 = 'z') @ output_density_10)
        expect_DQD_2_ex_2_read = np.trace(-operator_disp(spin_2 = 'z') @ output_density_10)
        
        if result == "expectation":
            measurement_DQD_1_ex_1_read.append(((expect_DQD_1_ex_1_read+1)/2))
            measurement_DQD_1_ex_2_read.append(((expect_DQD_1_ex_2_read+1)/2))
            
            measurement_DQD_2_ex_1_read.append(((expect_DQD_2_ex_1_read+1)/2))
            measurement_DQD_2_ex_2_read.append(((expect_DQD_2_ex_2_read+1)/2))
        elif result == "resonator":
            chi_DQD_1 = (1/(f_qubit_dot_1-res.f_read-1j*(g2_dot_1_spin+g1_dot_1_spin/2))
                    +1/(f_qubit_dot_1+res.f_read+1j*(g2_dot_1_spin+g1_dot_1_spin/2))
                    )
            # print(f" parameters for chi: {f_qubit_dot_1-res.f_read} {g2_dot_1_spin} {g1_dot_1_spin/2}")
            chi_DQD_2 = (1/(f_qubit_dot_2-res.f_read-1j*(g2_dot_2_spin+g1_dot_2_spin/2))
                    +1/(f_qubit_dot_2+res.f_read+1j*(g2_dot_2_spin+g1_dot_2_spin/2))
                    )
            # print(f" real part chi: {np.real(chi_DQD_1)}")
            # print(f" imag part chi: {np.imag(chi_DQD_1)}")
            measurement_DQD_1_ex_1_read_temp = 1j/(
                res.f_res - disp_shift_dot_1 - res.f_read - exchange*detuning*chi_DQD_1*expect_DQD_1_ex_1_read - 1j*res.k_tot/2
                )
            
            measurement_DQD_1_ex_2_read_temp = 1j/(
                res.f_res - disp_shift_dot_1 - res.f_read - exchange*detuning*chi_DQD_1*expect_DQD_1_ex_2_read - 1j*res.k_tot/2
                )
            
            measurement_DQD_2_ex_1_read_temp = 1j/(
                res.f_res - disp_shift_dot_2 - res.f_read - exchange*detuning*chi_DQD_2*expect_DQD_2_ex_1_read - 1j*res.k_tot/2
                )
            
            measurement_DQD_2_ex_2_read_temp = 1j/(
                res.f_res - disp_shift_dot_2 - res.f_read - exchange*detuning*chi_DQD_2*expect_DQD_2_ex_2_read - 1j*res.k_tot/2
                )
            
            renormalization_DQD_1 =  np.abs(1j/(
                res.f_res - disp_shift_dot_1 - res.f_read - exchange*detuning*chi_DQD_1 + 1j*kappa_res/2
                ))
            
            renormalization_DQD_2 =  np.abs(1j/(
                res.f_res - disp_shift_dot_1 - res.f_read - exchange*detuning*chi_DQD_2 + 1j*kappa_res/2
                ))
            
            measurement_DQD_1_ex_1_read.append(measurement_DQD_1_ex_1_read_temp/renormalization_DQD_1)
            measurement_DQD_1_ex_2_read.append(measurement_DQD_1_ex_2_read_temp/renormalization_DQD_1)
            
            measurement_DQD_2_ex_1_read.append(measurement_DQD_2_ex_1_read_temp/renormalization_DQD_2)
            measurement_DQD_2_ex_2_read.append(measurement_DQD_2_ex_2_read_temp/renormalization_DQD_2)
        
    if q_excite == 1:
        return np.array([measurement_DQD_1_ex_1_read, measurement_DQD_1_ex_2_read])
    elif q_excite == 2:
        return np.array([measurement_DQD_2_ex_1_read, measurement_DQD_2_ex_2_read])
    
# 

def plot_1d_experiment():
    
    time = np.linspace(0,120*1e-3,201)
    data = fitting_iswap_dispersive(time, result = 'expectation',
                                    f_qubit_dot_1 = 6779.2,
                                    f_qubit_dot_2 = 6779.2)
    
    data_dqd1_01 = data[0]
    data_dqd1_10 = data[1]
    
    
    plt.figure(figsize=(5, 4))
    plt.plot(time,np.abs(data_dqd1_10), 'bx-')
    plt.plot(time,np.abs(data_dqd1_10), 'gx-')
    plt.xlabel("time (microseconds)")
    plt.ylabel("probability")
    plt.legend(['data_dqd_1_ex_1_read','data_dqd_1_ex_2_read'])
    

def data_import(subfolder = 'YOURPATH/data/', figure: str ='3a'):
    
    

    if figure == '3b':
        data1 = np.load(subfolder+'Flip_Q1_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '3d':
        data1 = np.load(subfolder+'Flip_Q2_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '4a':
        data1 = np.load(subfolder+'Flip_Q1_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_22_MHz.npy')
        time_end = 160*1e-3
    elif figure == '4b':
        data1 = np.load(subfolder+'Flip_Q2_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_22_MHz.npy')  
        time_end = 160*1e-3
    elif figure == '4c':
        data1 = np.load(subfolder+'Flip_Q1_Q1_18_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_18_MHz.npy')  
        time_end = 120*1e-3
    elif figure == '4d':
        data1 = np.load(subfolder+'Flip_Q2_Q1_18_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_18_MHz.npy')    
        time_end = 120*1e-3
    


    return np.array([data1, data2])

def plot_results(q_excite: int = '1',figure: str ='3b', subfolder = 'YOURPATH/data/', list_parameters = np.zeros([9]), 
                 mode : str = 'expectation'):
    
    
    
    if figure == '3b':
        data1 = np.load(subfolder+'Flip_Q1_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '3d':
        data1 = np.load(subfolder+'Flip_Q2_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '4a':
        data1 = np.load(subfolder+'Flip_Q1_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_22_MHz.npy')
        time_end = 160*1e-3
    elif figure == '4b':
        data1 = np.load(subfolder+'Flip_Q2_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_22_MHz.npy')  
        time_end = 160*1e-3
    elif figure == '4c':
        data1 = np.load(subfolder+'Flip_Q1_Q1_18_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_18_MHz.npy')  
        time_end = 120*1e-3
    elif figure == '4d':
        data1 = np.load(subfolder+'Flip_Q2_Q1_18_MHz.npy')
        data2= np.load(subfolder+'Flip_Q2_Q2_18_MHz.npy')    
        time_end = 120*1e-3
    
    
    time = np.linspace(0,time_end,len(data1))
    timefit = np.linspace(0,time_end,1000)
    
    data = fitting_iswap_dispersive(timefit,exchange=list_parameters[2],
                                    result = mode,
                                    g1_dot_1_spin = list_parameters[5],
                                    g2_dot_1_spin = list_parameters[6],
                                    g1_dot_2_spin = list_parameters[7],
                                    g2_dot_2_spin = list_parameters[8],
                                    q_excite = q_excite)
    
    plt.figure(figsize=(5, 4))    
    plt.plot(time,np.abs(data1), 'gx-')
    plt.plot(time,np.abs(data2), 'bx-')
    # fit1= (data[0])*list_parameters[0] + list_parameters[3]
    # fit2 = (data[1])*list_parameters[1] + list_parameters[4]
    
    fit1= (data[0])*list_parameters[0] + list_parameters[3]
    fit2 = (data[1])*list_parameters[1] + list_parameters[4]
    plt.plot(timefit,np.abs(fit1), 'g-') 
    plt.plot(timefit,np.abs(fit2), 'b-')
    # plt.plot(time_steps,proj_red_11)
    plt.xlabel("time (microseconds)")
    plt.ylabel("probability")
    plt.show()
    
    # plt.figure(figsize=(5, 4))    
    # plt.plot(time,data_dqd2_10, 'gx-')
    # plt.plot(time,data_dqd2_01, 'bx-')
    # plt.plot(time,(data[2])*list_parameters[2] + list_parameters[7], 'g-')
    # plt.plot(time,(data[3])*list_parameters[3] + list_parameters[8], 'b-')
    # # plt.plot(time_steps,proj_red_11)
    # plt.xlabel("time (microseconds)")
    # plt.ylabel("probability")
    # plt.legend(['data_dqd2_10','data_dqd2_01'])
    # plt.show()
    return [fit1, fit2, timefit]

def fit_experimental_data(mode : str = 'expectation', figure: str = '3b', q_excite: int = 1):
    data_fit = data_import(figure=figure)
    if figure == '4c' or figure =='4d':
        time = np.linspace(0,120*1e-3,len(data_fit[0]))
    else:
        time = np.linspace(0,160*1e-3,len(data_fit[0]))
    
    temp_data_fit_1 = data_fit[0:2,:].flatten()
    # temp_data_fit_2 = data_fit[2:4,:].flatten()
    data = np.array(temp_data_fit_1)
    # fit
    global iterator
    iterator = 0
    def fitting_function(x,a1,b1,e,a0,b0,g1_1,g2_1,g1_2,g2_2):
        global iterator
        print(f" run: {iterator} with guess {np.around(np.array([e,1/g1_1,1/g1_2,1/g2_1,1/g2_2]),6)}")
        # if iterator == 0:
        #     print(a1,b1,a0,b0)
        iterator += 1
        temp = np.abs(fitting_iswap_dispersive(x,exchange=e,
                                        result = mode,
                                        g1_dot_1_spin = g1_1,
                                        g2_dot_1_spin = g2_1,
                                        g1_dot_2_spin = g1_2,
                                        g2_dot_2_spin = g2_2,
                                        q_excite = q_excite
                                        ))
        temp[0,:] = (temp[0,:])*a1 + a0
        temp[1,:] = (temp[1,:])*b1 + b0
        # temp[2,:] = (temp[2,:])*c1 + c0
        # temp[3,:] = (temp[3,:])*d1 + d0
        
        temp_1 = temp[0:2,:].flatten()
        # temp_2 = temp[2:4,:].flatten()
        return np.array(temp_1)
    
    popt, pcov = sp.optimize.curve_fit(fitting_function,time,data,
                                       bounds=(
        [
        (np.max(data_fit[0,:])-np.min(data_fit[0,:]))/8,
        (np.max(data_fit[1,:])-np.min(data_fit[1,:]))/8,
        # (np.max(data_fit[2,:])-np.min(data_fit[2,:]))/2,
        # (np.max(data_fit[3,:])-np.min(data_fit[3,:]))/2,
        10,
        -np.max(data_fit[0,:])*2,
        -np.max(data_fit[1,:])*2,
        dot_1.gamma_rel_s/5,
        dot_1.gamma_dep_s/5,
        dot_2.gamma_rel_s/5,
        dot_2.gamma_dep_s/5
        ],
        [(np.max(data_fit[0,:])-np.min(data_fit[0,:]))*8,
        (np.max(data_fit[1,:])-np.min(data_fit[1,:]))*8,
        # (np.max(data_fit[2,:])-np.min(data_fit[2,:]))*2,
        # (np.max(data_fit[3,:])-np.min(data_fit[3,:]))*2,
        22,
        np.max(data_fit[0,:])*8,
        np.max(data_fit[1,:])*8,
        dot_1.gamma_rel_s*5,
        dot_1.gamma_dep_s*5,
        dot_2.gamma_rel_s*5,
        dot_2.gamma_dep_s*5
        ]))
    print(popt)
    perr = np.sqrt(np.diag(pcov))
    print(perr)
    data = plot_results(q_excite,figure,list_parameters = popt, mode = mode, )
    data = np.real(data)
    return data

if __name__ == '__main__':
    
    '''
    This script contains the fits of the data to the dispersive model described in the paper
    When the script is run it will fit the data of figure 3b, other figures can be reconstructed by uncommenting the respective block
    Note to add the correct path in line 328 and 361, there the data traces of the datapoints of the figures plotted in the paper are imported, for the reconstruction of these traces see the 'data_with_codes' folder in the repository
    '''
    
    
    data = fit_experimental_data(mode = 'expectation', figure='3b',q_excite=1)
    # data = fit_experimental_data(mode = 'expectation', figure='3d',q_excite=2)
    # data = fit_experimental_data(mode = 'expectation', figure='4a',q_excite=1)
    # data = fit_experimental_data(mode = 'expectation', figure='4b',q_excite=2)
    # data = fit_experimental_data(mode = 'expectation', figure='4c',q_excite=1)
    # data = fit_experimental_data(mode = 'expectation', figure='4d',q_excite=2)
