import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
import pickle
import math

from dataclasses import dataclass, field

from scipy.constants import elementary_charge, hbar, electron_mass, physical_constants, speed_of_light

from scipy.sparse.linalg import expm, expm_multiply
from scipy.sparse import save_npz, load_npz, csr_array, csc_array, diags, kron, kronsum, identity
from scipy.sparse.linalg import lobpcg, LinearOperator, inv, norm

iterator = 0

@dataclass(frozen=True)
class ph_con():
    mu_b : float = physical_constants['Bohr magneton in eV/T'][0]
    hb : float = hbar
    m0 : float = electron_mass
    ec : float = elementary_charge
    c : float = speed_of_light
    mhz2ev : float = 4.13566553853598*1e-9
@dataclass(frozen=True)
class res():
    n_photon : int = 3
    k_tot : float = 2.5 # MHz
    k_int : float = 1.5 # Mhz
    f_res : float = 6910.5 # MHz
    f_read : float = 6904.6 # MHz

@dataclass(frozen=True)
class dot_1():
    
    gamma_rel_s : float = 1/0.05 # MHz
    gamma_dep_s : float = 1/0.2 # MHz
    
    
    leverarm : float = 30*1e-3 # eV/V
@dataclass(frozen=True)
class dot_2():

    
    gamma_rel_s : float = 1/0.11 # MHz
    gamma_dep_s : float = 1/0.09 # MHz
    
    leverarm : float = 30*1e-3 # eV/V


entries : np.ndarray = np.array( [np.sqrt(it+1) for it in range(0,res.n_photon-1)] , dtype = float)

@dataclass(frozen=True)
class operators():
    i : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,1]],dtype = complex))
    x : np.ndarray = field(default_factory = lambda :np.array([[0,1],[1,0]],dtype = complex))
    y : np.ndarray = field(default_factory = lambda :np.array([[0,-1j],[1j,0]],dtype = complex))
    z : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,-1]],dtype = complex))
    pl : np.ndarray = field(default_factory = lambda :np.array([[0,1],[0,0]],dtype = complex))
    mi : np.ndarray = field(default_factory = lambda :np.array([[0,0],[1,0]],dtype = complex))
    pl_star : np.ndarray = field(default_factory = lambda :np.array([[1,1],[-1,-1]],dtype = complex)/2)
    mi_star : np.ndarray = field(default_factory = lambda :np.array([[1,-1],[1,-1]],dtype = complex)/2)
    p0 : np.ndarray = field(default_factory = lambda :np.array([[1,0],[0,0]],dtype = complex))
    p1 : np.ndarray = field(default_factory = lambda :np.array([[0,0],[0,1]],dtype = complex))
    
    ii : np.ndarray = field(default_factory = lambda :np.identity(4,dtype = complex))
    i4 : np.ndarray = field(default_factory = lambda :np.identity(4**2,dtype = complex))
    
    
    a : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1))
    ad : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1).T)
    ada : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1).T @ np.diag(entries,k=1))
    # ada : np.ndarray = field(default_factory = lambda :np.diag(entries,k=1).T @ np.diag(entries,k=1))
    ire : np.ndarray = field(default_factory = lambda :np.identity(res.n_photon,dtype = complex))
    
    res0 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
    # res0[0,0] =1
    if res.n_photon > 1:
        res1 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res1[1,1] =1
    if res.n_photon > 2:
        res2 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res2[2,2] =1
    if res.n_photon > 3:
        res3 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res3[3,3] =1
    if res.n_photon > 4:
        res4 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res4[4,4] =1
    if res.n_photon > 5:
        res5 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res5[5,5] =1
    if res.n_photon > 6:
        res6 : np.ndarray = field(default_factory = lambda :np.zeros([res.n_photon,res.n_photon], dtype = complex))
        # res6[6,6] =1

pauli = dict({'i':np.array([[1,0],[0,1]],dtype = complex),
              'x':np.array([[0,1],[1,0]],dtype = complex),
              'y':np.array([[0,-1j],[1j,0]],dtype = complex),
              'z':np.array([[1,0],[0,-1]],dtype = complex),
              '+':np.array([[0,1],[0,0]],dtype = complex),
              '-':np.array([[0,0],[1,0]],dtype = complex),
              '+*':np.array([[1,1],[-1,-1]],dtype = complex)/2,
              '-*':np.array([[1,-1],[1,-1]],dtype = complex)/2,
              '0':np.array([[1,0],[0,0]],dtype = complex),
              '1':np.array([[0,0],[0,1]],dtype = complex),
              })

pauli['x+'] = (pauli['i']+pauli['x'])/2
pauli['x-'] = (pauli['i']-pauli['x'])/2
pauli['y+'] = (pauli['i']+pauli['y'])/2
pauli['y-'] = (pauli['i']-pauli['y'])/2
pauli['z+'] = (pauli['i']+pauli['z'])/2
pauli['z-'] = (pauli['i']-pauli['z'])/2

caps = dict({'i':operators().ire,
              'a':operators().a,
              'ad':operators().ad,
              'ada':operators().ada})

for it, value in enumerate([v for k, v in operators().__dict__.items() if 'res' in k]):
    caps[str(it)] = value

def operator(spin_1 : str = 'i', spin_2 : str = 'i', res : str = 'i') -> np.ndarray:
    op_s1 = pauli[spin_1]
    op_s2 = pauli[spin_2]
    op_res = caps[res]
    
    matrix = np.kron(np.kron(op_s1,op_s2),op_res)
    
    return matrix
                          
# states = dict({'00':operators.dm00,
#                '01':operators.dm01,
#                '10':operators.dm10,
#                '11':operators.dm11})


def compute_eigenvalues(ham, return_vecs : bool = False):
    # returns eigenvalues and optionally eigenvectors
    eig_val_s, eig_vec = np.linalg.eig(ham)
    idx = eig_val_s.argsort()
    eig_val = eig_val_s[idx]
    if return_vecs is False:
        return eig_val/(2*np.pi)
    else:
        eig_vec = eig_vec[:, idx].T
        return eig_val/(2*np.pi) , eig_vec
    

def fitting_iswap(t_array : np.ndarray, 
                  f_qubit_dot_1 : float = 6780.5, # total resonance frequency minus dispersive shift
                  f_qubit_dot_2 : float = 6780.5, # total resonance frequency minus dispersive shift
                  gs_dot_1 : float = 32., # spin-photon coupling dot 1
                  gs_dot_2 : float = 32., # spin-photon coupling dot 2
                  charge_disp_shift_dot_1 : float = 20.5 , # dispersive shift dot 1
                  charge_disp_shift_dot_2 : float = 20.5 , # dispersive shift dot 2
                  g1_dot_1_spin : float = dot_1.gamma_rel_s, # spin-relaxation dot 1
                  g2_dot_1_spin : float = dot_1.gamma_dep_s, # spin-dephasing dot 1
                  g1_dot_2_spin : float = dot_2.gamma_rel_s, # spin-relaxation dot 2
                  g2_dot_2_spin : float = dot_2.gamma_dep_s, # spin-dephasing dot 2
                  kappa_res : float = res.k_tot,
                  result : str = "expectation", # type of result: 
                                               #'expectation' normalized expectation values, 
                                               #'resonator' unnormalized readout values
                  q_excite : int = 1
                  ):
    # convert into MHz and set hbar = 1
    
    
    # Set up the Hamiltonian without coupling to get initial states
    hamiltonian_static = 2*np.pi*(
        # zero order terms
        + operator(res = 'ada') * (res.f_res)
        - operator(spin_1 = 'z') * (f_qubit_dot_1)/2
        - operator(spin_2 = 'z') * (f_qubit_dot_2)/2
        )
    
    # Set up time-steps. Currently only uniform time traces are supported
    t_array = t_array
    t_step = t_array[1]-t_array[0]
    # Set up the initial states from diagonalizing the non-interacting Hamiltonian
    t_eigs, t_vecs = compute_eigenvalues(hamiltonian_static, return_vecs = True)
    init_DQD1 = t_vecs[2]
    init_DQD2 = t_vecs[1]
    init_ground = t_vecs[0]
    
    
    # Set up the interacting Hamiltonian for dot 1 (dot 2 is not interacting with the resonator)
    hamiltonian_DQD_1 = (hamiltonian_static 
                         - 2*np.pi*(
                                 operator(spin_1 = 'x',res = 'a')
                                 +operator(spin_1 = 'x',res = 'ad')
                                    ) * gs_dot_1
                         - 2*np.pi*(operator(res = 'ada') * charge_disp_shift_dot_1)
                        )
    # Set up the interacting Hamiltonian for dot 2 (dot 1 is not interacting with the resonator)
    hamiltonian_DQD_2 = (hamiltonian_static 
                         - 2*np.pi*(
                                 operator(spin_2 = 'x',res = 'a')
                                 +operator(spin_2 = 'x',res = 'ad')
                                    ) * gs_dot_2
                         - 2*np.pi*(operator(res = 'ada') * charge_disp_shift_dot_2)
                        )
    # Set up the interacting Hamiltonian for both dots
    hamiltonian = (hamiltonian_DQD_1 + hamiltonian_DQD_2 -hamiltonian_static)
    
    idmatrix = np.identity(len(hamiltonian), dtype = complex)
    
    id_vec = idmatrix.flatten()
    
    t_eigs_2, t_vecs_2 = compute_eigenvalues(hamiltonian, return_vecs = True)
    # print((t_eigs[3]-t_eigs[1],t_eigs[3]-t_eigs[2]))
    # print((t_eigs_2[3]-t_eigs_2[1],t_eigs_2[3]-t_eigs_2[2]))
    print(t_eigs_2[2]-t_eigs_2[1])
    
    # Set up Liuovillian for the system
    liuoville = -1j*(np.kron(hamiltonian,idmatrix)
                            -np.kron(idmatrix,hamiltonian.T))
    
    
    # set up all Lindblad terms
    
    lindblad_1 = np.sqrt(g1_dot_1_spin) * operator(spin_1 = '+')
    lindblad_2 = np.sqrt(g1_dot_2_spin) * operator(spin_2 = '+')
    
    lindblad_3 = np.sqrt(g2_dot_1_spin/2) * operator(spin_1 = 'z')
    lindblad_4 = np.sqrt(g2_dot_2_spin/2) * operator(spin_2 = 'z')
    
    lindblad_5 = np.sqrt(kappa_res) * operator(res = 'a')
    
    
    # create lindbladadian
    lindblad = np.zeros(liuoville.shape,dtype=complex)
    
    list_lindblads = [lindblad_1,lindblad_2,lindblad_3,
                      lindblad_4,lindblad_5]
    
    # add Lindblad dissipators
    
    for dissipator in list_lindblads:
        lindblad += 1/2*(2*np.kron(dissipator,dissipator)
                          -np.kron(idmatrix,(dissipator.T @ dissipator).T)
                          -np.kron((dissipator.T @ dissipator),idmatrix))
    
    # create initial dm states
    init_dm = [np.outer(dm_a, dm_a.conj()) 
               for dm_a in [init_DQD1,init_DQD2,init_ground]]
    
    
    # vectorize initial dm states
    init_vec = [dm.flatten() for dm in init_dm]
    
    # Set up full system 
    lindbladian = lindblad + liuoville
    # Set up full system for a time step
    lindbladian_t = csc_array(lindbladian*t_step)
    
    # Set up state for each time step
    final_vec_01 = [init_vec[0]]
    final_vec_10 = [init_vec[1]]
    
    # Compute state for each time step by serial propagating it with lindbladian_t
    for iterator in range(len(t_array)-1):
        final_vec_01.append(expm_multiply(lindbladian_t,final_vec_01[-1]))
        final_vec_10.append(expm_multiply(lindbladian_t,final_vec_10[-1]))
    
    measurement_DQD_1_ex_1_read = []
    measurement_DQD_1_ex_2_read = []
    measurement_DQD_2_ex_1_read = []
    measurement_DQD_2_ex_2_read = []
    
    if result == 'expectation':
        
        
        for iterator in range(len(t_array)):
            
            output_density_01 = np.reshape(final_vec_01[iterator],[len(hamiltonian),len(hamiltonian)])
            output_density_10 = np.reshape(final_vec_10[iterator],[len(hamiltonian),len(hamiltonian)])
            
            expect_DQD_1_ex_1_read = np.trace(-operator(spin_1 = 'z') @ output_density_01)
            expect_DQD_1_ex_2_read = np.trace(-operator(spin_2 = 'z') @ output_density_01)
            
            expect_DQD_2_ex_1_read = np.trace(-operator(spin_1 = 'z') @ output_density_10)
            expect_DQD_2_ex_2_read = np.trace(-operator(spin_2 = 'z') @ output_density_10)
            
            measurement_DQD_1_ex_1_read.append((expect_DQD_1_ex_1_read+1)/2)
            measurement_DQD_1_ex_2_read.append((expect_DQD_1_ex_2_read+1)/2)
            
            measurement_DQD_2_ex_1_read.append((expect_DQD_2_ex_1_read+1)/2)
            measurement_DQD_2_ex_2_read.append((expect_DQD_2_ex_2_read+1)/2)
        
    elif result == 'resonator':
        # Set up coupling Hamiltonians for the readout. Theory can be found in 
        # Bonsen et all, PRL 130, 137001 10.1103/PhysRevLett.130.137001
        a_operator = operator(res = 'a')
        ad_operator = operator(res = 'ad') 
        
        
        liuoville_DQD_1 = -1j*(np.kron(hamiltonian_DQD_1,idmatrix)
                                -np.kron(idmatrix,hamiltonian_DQD_1.T))
        
        liuoville_DQD_2 = -1j*(np.kron(hamiltonian_DQD_2,idmatrix)
                                -np.kron(idmatrix,hamiltonian_DQD_2.T))
        # eigs, vecs = compute_eigenvalues(hamiltonian_DQD_1, return_vecs = True)
        
        
        
        output_vector_DQD_1 =  1j*id_vec @ (
            (np.kron(a_operator,idmatrix))
            @ np.linalg.inv(-liuoville_DQD_1-lindblad - 2j*np.pi*res.f_read*np.identity(len(lindbladian)))
            @ (np.kron(ad_operator,idmatrix)-np.kron(idmatrix,ad_operator.T))
            )
        output_vector_DQD_2 =  1j*id_vec @ (
            (np.kron(a_operator,idmatrix))
            @ np.linalg.inv(-liuoville_DQD_2-lindblad - 2j*np.pi*res.f_read*np.identity(len(lindbladian)))
            @ (np.kron(ad_operator,idmatrix)-np.kron(idmatrix,ad_operator.T))
            )
        # Create reference measurement
        
        renormalization_DQD_1 = output_vector_DQD_1 @ init_vec[-1]
        
        renormalization_DQD_2 = output_vector_DQD_2 @ init_vec[-1]
        
        
        # Simulate measurement using general input-output formalism
        
        for iterator in range(len(t_array)):
            
            measurement_DQD_1_ex_1_read.append((output_vector_DQD_1 @ final_vec_01[iterator]-renormalization_DQD_1)/np.abs(renormalization_DQD_1))
            measurement_DQD_1_ex_2_read.append((output_vector_DQD_2 @ final_vec_01[iterator]-renormalization_DQD_2)/np.abs(renormalization_DQD_2))
            
            measurement_DQD_2_ex_1_read.append((output_vector_DQD_1 @ final_vec_10[iterator]-renormalization_DQD_1)/np.abs(renormalization_DQD_1))
            measurement_DQD_2_ex_2_read.append((output_vector_DQD_2 @ final_vec_10[iterator]-renormalization_DQD_2)/np.abs(renormalization_DQD_2))
            
    if q_excite == 1:
        return np.array([measurement_DQD_1_ex_1_read, measurement_DQD_1_ex_2_read])
    elif q_excite ==2:
        return np.array([measurement_DQD_2_ex_1_read, measurement_DQD_2_ex_2_read])

def data_import(subfolder = 'YOURPATH/data/'):
    
    
    
    
    Flip_Q1_Q1_18_MHz = np.load(subfolder+'Flip_Q1_Q1_17_MHz.npy')
    Flip_Q1_Q2_18_MHz = np.load(subfolder+'Flip_Q1_Q2_17_MHz.npy')
    Flip_Q2_Q1_18_MHz = np.load(subfolder+'Flip_Q2_Q1_17_MHz.npy')
    Flip_Q2_Q2_18_MHz = np.load(subfolder+'Flip_Q2_Q2_17_MHz.npy')
    Flip_Q1_Q1_22_MHz = np.load(subfolder+'Flip_Q1_Q1_22_MHz.npy')
    Flip_Q1_Q2_22_MHz = np.load(subfolder+'Flip_Q1_Q2_22_MHz.npy')
    Flip_Q2_Q1_22_MHz = np.load(subfolder+'Flip_Q2_Q1_22_MHz.npy')
    Flip_Q2_Q2_22_MHz = np.load(subfolder+'Flip_Q2_Q2_22_MHz.npy')
    Flip_Q1_Q1_11_MHz = np.load(subfolder+'Flip_Q1_Q1_11_MHz.npy')
    Flip_Q1_Q2_11_MHz = np.load(subfolder+'Flip_Q1_Q2_11_MHz.npy')
    Flip_Q2_Q1_11_MHz = np.load(subfolder+'Flip_Q2_Q1_11_MHz.npy')
    Flip_Q2_Q2_11_MHz = np.load(subfolder+'Flip_Q2_Q2_11_MHz.npy')

    
    
    data_dqd1_10 = Flip_Q1_Q1_18_MHz
    data_dqd2_10 = Flip_Q1_Q2_18_MHz
    data_dqd1_01 = Flip_Q2_Q1_18_MHz
    data_dqd2_01 = Flip_Q2_Q2_18_MHz
    return np.array([data_dqd1_10, data_dqd2_10])

def plot_results(subfolder = 'YOURPATH/data/', list_parameters = np.zeros([9]), 
                 mode : str = 'resonator', datapoints : int = 1000, qubit_freq : float =6780.5, disp_shift : float = 20.5, figure: str = '9a', q_excite: int =1,
                 q_average : bool = False):
    
    
    if figure == '9a':
        data1 = np.load(subfolder+'Flip_Q1_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '9b':
        data1 = np.load(subfolder+'Flip_Q2_Q1_11_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_11_MHz.npy')
        time_end = 160*1e-3
    elif figure == '9c':
        data1 = np.load(subfolder+'Flip_Q1_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_22_MHz.npy')
        time_end = 160*1e-3
    elif figure == '9d':
        data1 = np.load(subfolder+'Flip_Q2_Q1_22_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_22_MHz.npy')  
        time_end = 160*1e-3
    elif figure == '9e':
        data1 = np.load(subfolder+'Flip_Q1_Q1_18_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q1_Q2_18_MHz.npy')  
        time_end = 120*1e-3
    elif figure == '9f':
        data1 = np.load(subfolder+'Flip_Q2_Q1_18_MHz.npy')
        data2 = np.load(subfolder+'Flip_Q2_Q2_18_MHz.npy')    
        time_end = 120*1e-3
    


    
    time = np.linspace(0,time_end,len(data1))
    timefit = np.linspace(0,time_end,datapoints)
    if q_average is False:
        data = fitting_iswap(timefit,gs_dot_1=list_parameters[2],
                             gs_dot_2=list_parameters[2],
                                        result = mode,
                                        f_qubit_dot_1 = qubit_freq,
                                        f_qubit_dot_2 = qubit_freq,
                                        charge_disp_shift_dot_1 = disp_shift, 
                                        charge_disp_shift_dot_2 = disp_shift, 
                                        g1_dot_1_spin = list_parameters[5],
                                        g2_dot_1_spin = list_parameters[6],
                                        g1_dot_2_spin = list_parameters[7],
                                        g2_dot_2_spin = list_parameters[8],
                                        q_excite = q_excite)

    
    plt.figure(figsize=(5, 4))    
    plt.plot(time,np.abs(data1), 'gx-')
    plt.plot(time,np.abs(data2), 'bx-')
    # fit1= (data[0])*list_parameters[0] + list_parameters[3]
    # fit2 = (data[1])*list_parameters[1] + list_parameters[4]
    
    fit1= (data[0])*list_parameters[0] + list_parameters[3]
    fit2 = (data[1])*list_parameters[1] + list_parameters[4]
    plt.plot(timefit,np.abs(fit1), 'g-') 
    plt.plot(timefit,np.abs(fit2), 'b-')
    # plt.plot(time_steps,proj_red_11)
    plt.xlabel("time (microseconds)")
    plt.ylabel("probability")
    plt.show()
    

    return [fit1, fit2, timefit]

def fit_experimental_data(mode : str = 'resonator'):
    data_fit = data_import()
    time = np.linspace(0,120*1e-3,len(data_fit[0]))
    temp_data_fit_1 = data_fit[0:2,:].flatten()
    # temp_data_fit_2 = data_fit[2:4,:].flatten()
    data = np.array(temp_data_fit_1)
    # fit
    global iterator
    iterator = 0
    def fitting_function(x,a1,b1,e,a0,b0,g1_1,g2_1,g1_2,g2_2):
        global iterator
        print(f" run: {iterator} with guess [T1_1,T1_2,T2_1,T2_2] {np.around(np.array([e,1/g1_1,1/g1_2,1/g2_1,1/g2_2]),6)}")
        
        iterator += 1
        temp = np.abs(fitting_iswap(x,gs_dot_1=e,
                                    gs_dot_2=e,
                                        result = mode,
                                        f_qubit_dot_1 = 6780.5,
                                        f_qubit_dot_2 = 6780.5,
                                        g1_dot_1_spin = g1_1,
                                        g2_dot_1_spin = g2_1,
                                        g1_dot_2_spin = g1_2,
                                        g2_dot_2_spin = g2_2
                                        ))
        temp[0,:] = (temp[0,:])*a1 + a0
        temp[1,:] = (temp[1,:])*b1 + b0
        # temp[2,:] = (temp[2,:])*c1 + c0
        # temp[3,:] = (temp[3,:])*d1 + d0
        
        temp_1 = temp[0:2,:].flatten()
        # temp_2 = temp[2:4,:].flatten()
        return np.array(temp_1)
    
    
    popt, pcov = sp.optimize.curve_fit(fitting_function,time,data,
                                       p0 = [1.08302657e-03, 5.70984695e-04, 3.09727789e+01, 1.39639374e-04,
                                             1.65113870e-04, 1.86473276e+01, 3.06230851e+01, 1.81818182e+00,
                                             2.66621141e+00],
                                       bounds=(
        [
        (np.max(data_fit[0,:])-np.min(data_fit[0,:]))/5,
        (np.max(data_fit[1,:])-np.min(data_fit[1,:]))/5,
        # (np.max(data_fit[2,:])-np.min(data_fit[2,:]))/2,
        # (np.max(data_fit[3,:])-np.min(data_fit[3,:]))/2,
        20,
        -np.max(data_fit[0,:])*2,
        -np.max(data_fit[1,:])*2,
        dot_1.gamma_rel_s/10,
        dot_1.gamma_dep_s/10,
        dot_2.gamma_rel_s/10,
        dot_2.gamma_dep_s/10
        ],
        [(np.max(data_fit[0,:])-np.min(data_fit[0,:]))*5,
        (np.max(data_fit[1,:])-np.min(data_fit[1,:]))*5,
        # (np.max(data_fit[2,:])-np.min(data_fit[2,:]))*2,
        # (np.max(data_fit[3,:])-np.min(data_fit[3,:]))*2,
        50,
        np.max(data_fit[0,:])*4,
        np.max(data_fit[1,:])*4,
        dot_1.gamma_rel_s*10,
        dot_1.gamma_dep_s*10,
        dot_2.gamma_rel_s*10,
        dot_2.gamma_dep_s*10
        ]))
    print(popt)
    perr = np.sqrt(np.diag(pcov))
    print(perr)
    data = plot_results(list_parameters = popt, mode = mode)
    data = np.real(data)
    return data

if __name__ == '__main__':
    
    '''
    This script contains the full model as well as the fits of the measurement data to the full model described in the paper
    The fitted parameters ('list_parameters') for the different figure panels are shown below, with the input parameters all extracted from measurement
    Figure 9a is uncommented and will be shown when the script is run, for the other panels one can uncomment the respective block
    Note to add the correct path in line 341 and 367, there the data traces of the datapoints of the figures plotted in the paper are imported, for the reconstruction of these traces see the 'data_with_codes' folder in the repository
    '''
    
    #Figure 9a
    plot_results(list_parameters = [1.02627431e-03, 9.18307782e-04, 2.09667729e+01, 1.37266703e-04,
      1.69219210e-04, 1.25000855e+01, 4.80795826e+00, 9.69300833e-01,
      6.42227398e+00], mode = 'resonator', datapoints=201,qubit_freq =6818, disp_shift = 13.5, figure='9a', q_excite =1, q_average = False)
    
    #Figure 9b
    # plot_results(list_parameters = [4.00726761e-04, 1.00131233e-03, 2.10185904e+01, 2.05826224e-04,
    #   2.02475862e-04, 2.00041569e+00, 5.29314296e-01, 1.70421074e+01,
    #   1.15733615e+00], mode = 'resonator', datapoints=201,qubit_freq =6818, disp_shift = 13.5, figure='9b', q_excite =2, q_average = False)
    
    #Figure 9c
    # plot_results(list_parameters = [1.23537681e-03, 4.30755182e-04, 2.98641017e+01, 9.88698640e-05,
    #   1.63135193e-04, 2.03278294e+01, 2.14111503e+01, 9.22598525e-01,
    #   4.47919146e+00], mode = 'resonator', datapoints=201,qubit_freq =6806, disp_shift = 20.5, figure='9c', q_excite =1, q_average = False)
    
    #Figure 9d
    # plot_results(list_parameters = [6.45490313e-04, 6.29168596e-04, 2.95325666e+01, 1.23825525e-04,
    #   7.53168109e-05, 2.01272015e+00, 3.10597169e+00, 1.63648012e+01,
    #   2.18630764e+01], mode = 'resonator', datapoints=201,qubit_freq =6806, disp_shift = 20.5, figure='9d',q_excite=2, q_average = False)
    
    #Figure 9e
    # plot_results(list_parameters = [1.08275882e-03, 5.69374948e-04, 3.09041701e+01, 1.50421786e-04,
    #   1.70161442e-04 ,2.11048380e+01, 1.29744806e+01 ,9.16008440e-01,
    #   8.13366196e+00], mode = 'resonator', datapoints=201, qubit_freq =6780.5, disp_shift = 20.5, figure='9e',q_excite = 1,q_average = False)
    
    #Figure 9f
    # plot_results(list_parameters = [5.93587106e-04 ,8.58962562e-04, 3.15971463e+01 ,1.44578271e-04,
    #   8.56232641e-05, 2.00736212e+00 ,5.07646094e+00 ,1.78699224e+01,
    #   1.05908352e+01], mode = 'resonator', datapoints=201, qubit_freq =6780.5, disp_shift = 20.5, figure='9f',q_excite=2, q_average = False)
    